/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.server.config;

import java.io.IOException;
import java.io.File;
import java.io.StringWriter;
import java.io.PrintWriter;

import java.net.NetworkInterface;
import java.util.Enumeration;
import java.net.InetAddress;
import java.net.*;

import java.util.Iterator;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.LogManager;

import sync4j.framework.config.*;
import sync4j.framework.engine.SyncStrategy;
import sync4j.framework.engine.pipeline.PipelineManager;
import sync4j.framework.security.Officer;
import sync4j.framework.server.session.SessionHandler;
import sync4j.framework.server.store.PersistentStore;
import sync4j.framework.tools.beans.BeanFactory;
import sync4j.framework.tools.beans.BeanException;

import sync4j.server.admin.UserManager;

/**
 * Incapsulates all the configuration information about the Sync4j Engine.
 * Configuration parameters are stored as properties in a Map structure. The
 * key is the name of the parameter and it is strucured in dotted separated
 * namespaces. The value of the parameter is the value of the key and can be of
 * any type. Accessor methods are provided to get the value of the parameter in
 * a particular type (such as String, int, double, boolean, ...).<br>
 * Access to getXXX() and setXXX() methods are synchronized.
 * <p>
 * This class follows the Singleton pattern, therefore it cannot be directly
 * instantiated; an internal instance is created just once and returned at any
 * getConfiguration() call.
 *
 * @version $Id: Configuration.java,v 1.12 2005/10/12 10:48:59 luigiafassina Exp $
 */
public class Configuration
implements java.io.Serializable, ConfigurationConstants {

    // ------------------------------------------------------------ Private data

    /**
     * The server configuration
     */
    private ServerConfiguration serverConfig;

    private static Configuration singleton;
    private transient ClassLoader classLoader = null;

    //
    // In order to have a logger completely independent by any configuration,
    // we use a standard anounymous logger. This will be used for error
    // notification only.
    //
    private static transient Logger log = Logger.getAnonymousLogger();
    
    // -------------------------------------------------------------- properties

    /** Getter for property classLoader.
     * @return Value of property classLoader.
     *
     */
    public ClassLoader getClassLoader() {
        return classLoader;
    }

    /** Setter for property classLoader.
     * @param classLoader New value of property classLoader.
     *
     */
    public void setClassLoader(ClassLoader classLoader) {
        this.classLoader = classLoader;
    }

    // ------------------------------------------------------------ Constructors

    /**
     * This is disabled for normal use since this class is a singleton.
     */
    protected Configuration() {
    }

    /**
     * Returns the singleton instance. It created the instance the first time it
     * is called.
     *
     * @return the singleton instance.
     */
    public static synchronized Configuration getConfiguration() {
        if (singleton == null) {
            singleton = new Configuration();
            try {
                URL configPath = ConfigTools.fixURL(getConfigPath());
                ConfigClassLoader cl = new ConfigClassLoader(
                                           new URL[] { configPath },
                                           Configuration.class.getClassLoader()

                                       );
                singleton.serverConfig =
                    (ServerConfiguration)ConfigTools.getBeanInstance(cl, BEAN_SERVER_CONFIGURATION);

                singleton.setClassLoader(cl);

                singleton.setLoggingConfiguration();

                singleton.setServerURI();

            } catch (Exception e) {
                String msg = "Fatal error creating the configuration object ";
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe(msg + e.getMessage());
                }
                log.throwing(Configuration.class.getName(), "getConfiguration", e);
                //
                // In this way the exception is logged only once because if the log
                // level is FINER then the exception is logged by Throwing.
                //
                if (log.isLoggable(Level.INFO) && !log.isLoggable(Level.FINER)) {
                    StringWriter sw = new StringWriter();
                    e.printStackTrace(new PrintWriter(sw));
                    log.info(sw.toString());
                }
            }
        }
        return singleton;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Getter for property serverConfig.
     * @return Value of property serverConfig.
     */
    public ServerConfiguration getServerConfig() {
        return serverConfig;
    }

    /**
     * Load the server configuration
     */
    public void loadServerConfig() {
        //
        // In order to load the server configuration, we call the static
        // synchronized getConfiguration method after we have set the singleton object
        // to null. So, the getConfiguration method reload the server configuration bean
        // and re-set all the properties depends on it
        //
        singleton = null;
        getConfiguration();
    }

    /**
     * Creates and returns a new SyncStrategy instance from the configured
     * <i>serverConfig.getEngineConfiguration().strategy</i> property
     *
     * @return the newly created instance
     */
    public SyncStrategy getStrategy() {
        try {
            return (SyncStrategy)ConfigTools.getBeanInstance(
                classLoader,
                serverConfig.getEngineConfiguration().getStrategy()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the SyncStrategy object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getStartegy", e);
            //
            // In this way the exception is logged only once because if the log
            // level is FINER then the exception is logged by Throwing.
            //
            if (log.isLoggable(Level.INFO) && !log.isLoggable(Level.FINER)) {
                StringWriter sw = new StringWriter();
                e.printStackTrace(new PrintWriter(sw));
                log.info(sw.toString());
            }
        }

        return null;
    }

    /**
     * Creates and returns a new PersistentStore instance from the configured
     * <i>serverConfig.getEngineConfiguration().store</i> property
     *
     * @return the newly created instance
     */
    public PersistentStore getStore() {
        try {
            return (PersistentStore)ConfigTools.getBeanInstance(
                classLoader,
                serverConfig.getEngineConfiguration().getStore()
            );
        } catch (Exception e) {                        
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the PersistentStore object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getStore", e);
            //
            // In this way the exception is logged only once because if the log
            // level is FINER then the exception is logged by Throwing.
            //
            if (log.isLoggable(Level.INFO) && !log.isLoggable(Level.FINER)) {
                StringWriter sw = new StringWriter();
                e.printStackTrace(new PrintWriter(sw));
                log.info(sw.toString());
            }
        }

        return null;
    }

    /**
     * Creates and returns a new Officer instance from the configured
     * <i>serverConfig.getEngineConfiguration().getOfficer</i> property
     *
     * @return the newly created instance
     */
    public Officer getOfficer() {
        try {
            return (Officer)ConfigTools.getBeanInstance(
                classLoader,
                serverConfig.getEngineConfiguration().getOfficer()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the SecurityOfficer object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getOfficer", e);
            //
            // In this way the exception is logged only once because if the log
            // level is FINER then the exception is logged by Throwing.
            //
            if (log.isLoggable(Level.INFO) && !log.isLoggable(Level.FINER)) {
                StringWriter sw = new StringWriter();
                e.printStackTrace(new PrintWriter(sw));
                log.info(sw.toString());
            }
        }

        return null;
    }


    /**
     * Creates and returns a new SessionHandler instance from the configured
     * <i>serverConfig.getEngineConfiguration().sessionHandler</i> property
     *
     * @return the newly created instance
     */
    public SessionHandler getSessionHandler() {
        try {
            return (SessionHandler)ConfigTools.getBeanInstance(
                classLoader,
                serverConfig.getEngineConfiguration().getSessionHandler()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the SessionHandler object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getSessionHandler", e);
            //
            // In this way the exception is logged only once because if the log
            // level is FINER then the exception is logged by Throwing.
            //
            if (log.isLoggable(Level.INFO) && !log.isLoggable(Level.FINER)) {
                StringWriter sw = new StringWriter();
                e.printStackTrace(new PrintWriter(sw));
                log.info(sw.toString());
            }
        }

        return null;
    }

    /**
     * Creates and returns a new PipelineManager instance from the configured
     * <i>serverConfig.getEngineConfiguration().pipelineManager</i> property
     *
     * @return the newly created instance
     */
    public PipelineManager getPipelineManager() {
        try {
            return (PipelineManager)ConfigTools.getBeanInstance(
                classLoader,
                serverConfig.getEngineConfiguration().getPipelineManager()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the PipelineManager object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getPipelineManager", e);
            //
            // In this way the exception is logged only once because if the log
            // level is FINER then the exception is logged by Throwing.
            //
            if (log.isLoggable(Level.INFO) && !log.isLoggable(Level.FINER)) {
                StringWriter sw = new StringWriter();
                e.printStackTrace(new PrintWriter(sw));
                log.info(sw.toString());
            }
        }

        return null;
    }

    /**
     * Creates and returns a new UserManager instance from the configured
     * <i>serverConfig.getEngineConfiguration().userManager</i> property
     *
     * @return the newly created instance
     */
    public UserManager getUserManager() {
        try {
            return (UserManager)ConfigTools.getBeanInstance(
                classLoader,
                serverConfig.getEngineConfiguration().getUserManager()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the UserManager object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getUserManager", e);
            //
            // In this way the exception is logged only once because if the log
            // level is FINER then the exception is logged by Throwing.
            //
            if (log.isLoggable(Level.INFO) && !log.isLoggable(Level.FINER)) {
                StringWriter sw = new StringWriter();
                e.printStackTrace(new PrintWriter(sw));
                log.info(sw.toString());
            }
        }

        return null;
    }

    /**
     * Creates and returns a new Logging instance from the configured
     * <i>serverConfig.getEngineConfiguration().logging</i> property
     *
     * @return the newly created instance
     */
    public LoggingConfiguration getLoggingConfiguration() {
        try {
            return (LoggingConfiguration)ConfigTools.getBeanInstance(
                classLoader,
                serverConfig.getEngineConfiguration().getLoggingConfiguration()
            );
        } catch (Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error creating the LoggingConfiguration object: " + e.getMessage());
            }
            log.throwing(getClass().getName(), "getLoggingConfiguration", e);
            //
            // In this way the exception is logged only once because if the log
            // level is FINER then the exception is logged by Throwing.
            //
            if (log.isLoggable(Level.INFO) && !log.isLoggable(Level.FINER)) {
                StringWriter sw = new StringWriter();
                e.printStackTrace(new PrintWriter(sw));
                log.info(sw.toString());
            }
        }
        return null;
    }

    /**
     * Configures the logging system accordingly to the
     * sync4j/server/logging/Logging.xml server bean.
     *
     * @throws IOException in case of errors setting up file handlers
     *
     */
    public void setLoggingConfiguration() throws IOException {

        //
        // First of all reinitialize logging
        //
        LogManager lm = LogManager.getLogManager();
        lm.reset();
        lm.readConfiguration();

        //
        // Now we can reconfigure Sync4j logging
        //
        Logger l;
        LoggerConfiguration conf;
        LoggingConfiguration lc = getLoggingConfiguration();

        Handler h;
        Iterator i = lc.getLoggers().iterator();
        while(i.hasNext()) {

            conf = (LoggerConfiguration)i.next();

            //
            // If this logger inherits from its parent, just skip logger
            // configuration.
            //
            if (conf.isInherit()) {
                continue;
            }

            l = Logger.getLogger(conf.getName());

            //
            // first remove all old handlers
            //
            Handler[] handlers = l.getHandlers();
            for (int j=0; ((handlers != null) && (j<handlers.length)); ++j) {
                handlers[j].flush();
                handlers[j].close();
                l.removeHandler(handlers[j]);
            }
            l.setUseParentHandlers(false);

            //
            // Now process the new handlers.
            // Note that at the handler stage we set level to ALL (use the
            // logger level to select the logging granularity.
            //
            if (conf.isConsoleOutput()) {
                h = new ConsoleHandler();
                h.setFormatter(new sync4j.framework.logging.SimpleFormatter());
                h.setLevel(Level.ALL);
                l.addHandler(h);
            }

            if (conf.isFileOutput()) {
                String pattern = conf.getPattern();

                if (new File(pattern).isAbsolute() == false) {
                    pattern = getSync4jHome() + File.separator + pattern;
                }

                if ((pattern == null) || (pattern.trim().length() == 0)) {
                    pattern = "logs/sync4j%t.log";
                }
                h = new FileHandler(
                    pattern        ,
                    conf.getLimit()*1048576,
                    conf.getCount(),
                    conf.isAppend()
                );
                h.setFormatter(new sync4j.framework.logging.SimpleFormatter());
                h.setLevel(Level.ALL);
                l.addHandler(h);
            }

            if (LoggerConfiguration.LEVEL_OFF.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.OFF);
            } else if (LoggerConfiguration.LEVEL_SEVERE.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.SEVERE);
            } else if (LoggerConfiguration.LEVEL_INFO.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.INFO);
            } else if (LoggerConfiguration.LEVEL_ALL.equalsIgnoreCase(conf.getLevel())) {
                l.setLevel(Level.ALL);
            }

        }
    }

    /**
     * Loads and instantiate a bean by its config name. In this case the bean
     * is not looked up in the internal configuration map, but is created
     * directly by the means of the BeanFactory. This call is a shortcut for:
     * <code>
     *
     *  ClassLoader cl = Configuration.getConfiguration().getClassLoader();
     *  Object o = BeanFactory.getBeanInstance(cl, beanName);
     *
     * </code>
     *
     * @param beanName the complete beanName
     *
     * @return the bean instance
     *
     * @throws BeanException in case of instantiation error
     */
    public Object getBeanInstanceByName(String beanName)
    throws BeanException {
        return BeanFactory.getBeanInstance(classLoader, beanName);
    }

    /**
     * Saves an instantiate of a server bean to the given name under the config 
     * path.
     *
     * @param beanName the complete beanName
     * @param obj the bean instance
     *
     *
     * @throws BeanException in case of instantiation error
     */
    public void setBeanInstance(String beanName, Object o)
    throws BeanException {
        BeanFactory.saveBeanInstance(o, new File(getConfigPath(), beanName));
    }
    
    /**
     * Returns the Sync4j home as set as follows (the first value found is
     * taken):
     * <ul>
     *  <li>the system property sync4j.home</li>
     *  <li>the current directory</i>
     * </ul>
     *
     * @return the Sync4j home as set as discribed above.
     */
    public static String getSync4jHome() {
        return System.getProperty(PROPERTY_SYSTEM_SYNC4J_HOME, ".");
    }

    /**
     * Returns the config path calculated as getSync4jHome() + CONFIG_PATH
     *
     * @return the config path calculated as getSync4jHome() + CONFIG_PATH
     */
    public static String getConfigPath() {
        return getSync4jHome() + CONFIG_PATH;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();

        sb.append(getClass().getName()).append('\n')
          .append("serverInfo: ").append(serverConfig.getServerInfo()).append('\n')
          .append("engineConfiguration: ").append(serverConfig.getEngineConfiguration()).append('\n')
          .append("classLoader: " + classLoader).append('\n');

        return sb.toString();
    }

    /**
     * The server URI is optional into Sync4j.properties: so if it's not given,
     * it should be set at default value http://<localhost>:8080/sync4j/sync
     * This is to avoid to change the configuration in order to access at the
     * server remotely. When server URI is specified, the given value must be
     * used instead of the default.
     */
    private void setServerURI() {
        String serverURI = null;
        try {
            serverURI =
                singleton.getServerConfig().getEngineConfiguration().getServerURI();
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Server URI set: " + serverURI);
            }

            if (serverURI == null || serverURI.equals("")) {

                if (log.isLoggable(Level.WARNING)) {
                    log.warning("serverURI property not provided; looking for local ip address.");
                }

                serverURI = "http://"
                          + getServerAddress()
                          + ":8080/sync4j/sync"
                          ;

                singleton.getServerConfig().getEngineConfiguration().setServerURI(serverURI);
            }
        } catch (Exception e) {
            String msg = "Error in setting default server URI";
            log.severe(msg + ": " + e.getMessage());
            log.throwing(Configuration.class.getName(), "setServerURI", e);
            //
            // In this way the exception is logged only once because if the log
            // level is FINER then the exception is logged by Throwing.
            //
            if (log.isLoggable(Level.INFO) && !log.isLoggable(Level.FINER)) {
                StringWriter sw = new StringWriter();
                e.printStackTrace(new PrintWriter(sw));
                log.info(sw.toString());
            }
            throw new ConfigurationException(msg);
        }

        if (log.isLoggable(Level.INFO)) {
            log.info("ServerURI: " + serverURI);
        }
    }

    /**
     * Get the server address checking all network interfaces. If there is an address
     * not loopback len 4 bytes return it, also return the host name. If a error
     * occurs getting the host name, returns <i>localhost</i>
     * @return String
     */
    private String getServerAddress() {
        NetworkInterface iface = null;
        InetAddress ia = null;
        String address = null;

        try {
            for (Enumeration ifaces = NetworkInterface.getNetworkInterfaces();
                                      ifaces.hasMoreElements(); ) {
                iface = (NetworkInterface)ifaces.nextElement();
                for (Enumeration ips = iface.getInetAddresses(); ips.hasMoreElements(); ) {
                    ia = (InetAddress)ips.nextElement();
                    if (!ia.isLoopbackAddress() && (ia.getAddress().length == 4)) {
                        address = ia.getHostAddress();

                        if (log.isLoggable(Level.WARNING)) {
                            log.warning("Local address <" + address + "> found on interface <" +
                                        iface + ">");
                        }
                        return address;
                    }

                }
            }
        } catch (SocketException ex) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe(ex.getMessage());
            }
            return null;
        }

        if (address == null) {
            if (log.isLoggable(Level.WARNING)) {
                log.warning("Non loopback address found, using local host name. " +
                            "If you want to access the server from an external network, please check your network " +
                            "configuration and/or provide a valid serverURI " +
                            "property in Sync4j.xml. For local access, ignore this message.");
            }

            try {
                address = java.net.InetAddress.getLocalHost().getHostName();
            } catch (UnknownHostException ex) {
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe(ex.getMessage());
                }
                return null;
            }
        }
        return address;
    }
}
